/******************************************************************************
*
* project name:  FAT-Engine
* file name:     tex2hdr.c
* initial date:  05/04/2002
* author:        thomas.nussbaumer@gmx.net
* description:   tool to convert a binary file which contains texture data
*                for the FAT engine into a pretty nice headerfile
*
* the textures in the binary file has to be aligned like the output of
* fixtex64 namely this way:
*
* for grayscales (2 planes):
*
* [tile 1 plane 0 data]
* [tile 1 plane 1 data]
* [tile 2 plane 0 data]
* [tile 2 plane 1 data]
* ..... and so on .....
*
* for mono (1 plane):
*
* [tile 1 data]
* [tile 2 plane 0 data]
* ..... and so on .....
*
* each plane 64x64 pixels large, thus 512 (64*8) large
*
*
* $Id: tex2hdr.c,v 1.2 2002/04/05 15:24:46 tnussb Exp $
*
******************************************************************************/
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#define CVS_TRUNC_PREFIX    ((strlen((CVS_FILE_REVISION))<=11) ? 0 : (CVS_FILE_REVISION+11))
#define CVS_FIND_COMMA      (strchr(CVS_TRUNC_PREFIX,'.'))
#define CVSREV_MAIN         (int)(!(CVS_TRUNC_PREFIX) ? 0 : atoi(CVS_TRUNC_PREFIX))
#define CVSREV_SUB          (int)(!(CVS_TRUNC_PREFIX) ? 0 : (!(CVS_FIND_COMMA) ? 0 : atoi(CVS_FIND_COMMA+1)))
#define CVSREV_PRINTPARAMS  "v%d.%02d",CVSREV_MAIN,CVSREV_SUB
#define PRINT_ID(name)      fprintf(stderr,""name" ");fprintf(stderr,CVSREV_PRINTPARAMS);\
                            fprintf(stderr," - Part of FAT-Engine\n" \
                                   "(c) thomas.nussbaumer@gmx.net "__DATE__" "__TIME__"\n");
#ifdef CVS_FILE_REVISION
#undef CVS_FILE_REVISION
#endif
#define CVS_FILE_REVISION "$Revision: 1.2 $"


//=============================================================================
// outputs usage information of this tool
//=============================================================================
void PrintUsage() {
    fprintf(stderr, "Usage: Tex2Hdr [options] <infile> <outfile> [hcount vcount]\n\n"      \
                    "       -gray       ... process as 64x64 grayscale tiles (default)\n"  \
                    "       -mono       ... process as 64x64 monochrome tiles\n"           \
                    "       -prefix str ... prefix for tile names in headerfile\n"         \
                    "       infile      ... input file to use (generated by FixTex64)\n"   \
                    "       outfile     ... output file to use (use \"-\" for stdout)\n\n" \
                    "       converts the output of FixTex64 to a C header file\n\n");
}

void ProcessTile(FILE* ifp, FILE* ofp,int tile_number, int plane, int nr_planes,char* prefix) {
    unsigned char buffer[512];
    int           i;
    int           j;

    fseek(ifp,tile_number*(512*nr_planes)+512*plane,SEEK_SET);
    fread(buffer,512,1,ifp);

    for (i=0;i<4;i++) {
         fprintf(ofp,"#define %s%d ",prefix,i);
         for (j=0;j<64;j++) {
            if (j<63) {
                if (j%8==0) fprintf(ofp,"\\\n  ");
                fprintf(ofp,"0x%02x%02x,",buffer[i*128+j*2],buffer[i*128+j*2+1]);
            }
            else {
                fprintf(ofp,"0x%02x%02x\n\n",buffer[i*128+j*2],buffer[i*128+j*2+1]);
            }
         }
    }
}


//=============================================================================
// its a main ...
//=============================================================================
int main(int argc,char *argv[]) {
    char* infile    = NULL;
    char* outfile   = NULL;
    char* prefix    = "TILE";
    FILE* ifp;
    FILE* ofp;
    long  n;
    int   nr_planes = 2;
    int   quiet     = 0;
    long  length;
    long  nr_tiles;
    long  plane;

    // parse arguments
    for (n=1; n<argc; n++) {
        if (!strcmp(argv[n], "-gray"))        nr_planes = 2;
        else if (!strcmp(argv[n], "-mono"))   nr_planes = 1;
        else if (!strcmp(argv[n], "-quiet"))  quiet     = 1;
        else if (!strcmp(argv[n], "-prefix")) {
            if (n == argc -1) {
                PRINT_ID("Tex2Hdr");
                PrintUsage();
                return 1;
            }
            else {
                n++;
                prefix = argv[n];
            }
        }
        else if (!infile)                    infile    = argv[n];
        else if (!outfile)                   outfile   = argv[n];
        else {
            PRINT_ID("Tex2Hdr");
            PrintUsage();
            return 1;
        }
    }

    if (outfile && !strcmp(outfile,"-")) quiet = 1;

    if (!infile || !outfile) {
        PrintUsage();
        return 1;
    }

    if (!(ifp = fopen(infile,"rb"))) {
        fprintf(stderr,"ERROR: cannot open inputfile %s\n",infile);
        return 1;
    }

    fseek(ifp,0,SEEK_END);
    length = ftell(ifp);
    rewind(ifp);

    nr_tiles = length/(512*nr_planes);

    if ((nr_tiles*(512*nr_planes)) != length) {
        fprintf(stderr,"ERROR: filesize (%ld) doesn't match to tilesize (%ld)\n",length,nr_tiles*(512*nr_planes));
        fclose(ifp);
        return 1;
    }

    if (!strcmp(outfile,"-")) {
        if (!quiet) fprintf(stderr,"using stdout as output\n");
        ofp = stdout;
    }
    else {
        if (!(ofp = fopen(outfile,"w"))) {
            fprintf(stderr,"ERROR: cannot open ouputfile %s\n",outfile);
            fclose(ifp);
            return 1;
        }
    }

    for (n=0;n<nr_tiles;n++) {
       for (plane=0;plane<nr_planes;plane++) {
           char tmpstr[1024];
           sprintf(tmpstr,"%sTILE%dPLANE%dSTRIP",prefix,n,plane);
           ProcessTile(ifp,ofp,n,plane,nr_planes,tmpstr);
       }
    }

    fclose(ifp);
    fclose(ofp);
    if (!quiet) fprintf(stderr,"%ld tiles sucessfully processed.\n\n",nr_tiles);
    return 0;
}

//=============================================================================
// Revision History
//=============================================================================
//
// $Log: tex2hdr.c,v $
// Revision 1.2  2002/04/05 15:24:46  tnussb
// naming convention changed
//
// Revision 1.1  2002/04/05 15:08:38  tnussb
// initial version
//
//
//
